from collections.abc import Callable

from torch import nn


def get_NormalInitializer(std: float) -> Callable[[nn.Module], nn.Module]:
    def initializer(m: nn.Module) -> nn.Module:
        if isinstance(m, nn.Linear):
            nn.init.normal_(m.weight, 0, std)
            nn.init.normal_(m.bias, 0, std)

    return initializer



from __future__ import annotations

from abc import ABCMeta
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import fields
from typing import Any

import torch
from torch.utils.data import DataLoader


@dataclass
class Batch:
    """
    A batch of data, with non-optional x, y, and target_y attributes.
    All other attributes are optional.

    If you want to add an attribute for testing only, you can just assign it after creation like:
    ```
        batch = Batch(x=x, y=y, target_y=target_y)
        batch.test_attribute = test_attribute
    ```
    """

    # Required entries
    x: torch.Tensor
    y: torch.Tensor
    target_y: torch.Tensor

    # Optional Batch Entries
    style: torch.Tensor | None = None
    style_hyperparameter_values: torch.Tensor | None = None
    single_eval_pos: torch.Tensor | None = None
    causal_model_dag: object | None = None
    mean_prediction: bool | None = (
        None  # this controls whether to do mean prediction in bar_distribution for nonmyopic BO
    )

    def other_filled_attributes(
        self, set_of_attributes: set[str] = set(("x", "y", "target_y"))
    ) -> list[str]:
        return [
            f.name
            for f in fields(self)
            if f.name not in set_of_attributes and getattr(self, f.name) is not None
        ]


def safe_merge_batches_in_batch_dim(*batches: Any, ignore_attributes: list[str] = []) -> Batch:
    """
    Merge all supported non-None fields in a pre-specified (general) way,
    e.g. mutliple batch.x are concatenated in the batch dimension.
    :param ignore_attributes: attributes to remove from the merged batch, treated as if they were None.
    :return:
    """
    not_none_fields = [
        f.name
        for f in fields(batches[0])
        if f.name not in ignore_attributes and getattr(batches[0], f.name) is not None
    ]
    assert all(
        [
            set(not_none_fields)
            == set(
                [
                    f.name
                    for f in fields(b)
                    if f.name not in ignore_attributes and getattr(b, f.name) is not None
                ]
            )
            for b in batches
        ]
    ), "All batches must have the same fields!"
    merge_funcs = {
        "x": lambda xs: torch.cat(xs, 1),
        "y": lambda ys: torch.cat(ys, 1),
        "target_y": lambda target_ys: torch.cat(target_ys, 1),
        "style": lambda styles: torch.cat(styles, 0),
    }
    assert all(
        f in merge_funcs for f in not_none_fields
    ), "Unknown fields encountered in `safe_merge_batches_in_batch_dim`."
    return Batch(
        **{f: merge_funcs[f]([getattr(batch, f) for batch in batches]) for f in not_none_fields}  # type: ignore
    )


def merge_batches(*batches: Any, ignore_attributes: list[str] = []) -> Batch:
    assert False, "TODO: isn't this broken!? because catting in dim 0 seems wrong!?"

    def merge_attribute(attr_name: str, batch_sizes: list[int]) -> Any:
        attr = [getattr(batch, attr_name) for batch in batches]
        if isinstance(attr[0], list):

            def make_list(sublist, i):
                if sublist is None:
                    return [None for _ in range(batch_sizes[i])]
                return sublist

            return sum([make_list(sublist, i) for i, sublist in enumerate(attr)], [])
        elif type(attr[0]) is torch.Tensor:
            return torch.cat(attr, 0)
        else:
            assert all(a is None for a in attr), (
                f"Unknown type encountered in `merge_batches`."
                f"To ignore this, please add `{attr}` to the `ignore_attributes`."
                f"The following values are the problem: {attr_name}."
            )
            return None

    batch_sizes = [batch.x.shape[0] for batch in batches]
    return Batch(
        **{
            f.name: merge_attribute(f.name, batch_sizes)
            for f in fields(batches[0])
            if f.name not in ignore_attributes
        }
    )


class PriorDataLoader(DataLoader, metaclass=ABCMeta):
    @abstractmethod
    def __init__(
        self,
        num_steps: int,
        batch_size: int,
        eval_pos_seq_len_sampler: Callable,
        seq_len_maximum: int,
        device: torch.device | None,
        **kwargs: Any,
    ) -> None:
        """

        :param num_steps: int, first argument, the number of steps to take per epoch, i.e. iteration of the DataLoader
        :param batch_size: int, number of datasets per batch
        :param eval_pos_seq_len_sampler: callable, it takes no arguments and returns a tuple (single eval pos, bptt)
        :param kwargs: for future compatibility it is good to have a final all catch, as new kwargs might be introduced
        """
        pass

    # A class or object variable `num_features`: int
    # Optional: `validate` function that accepts a transformer model

    # The DataLoader iter should return batches of the form ([style], x, y), target_y, single_eval_pos
    # We follow sequence len (s) first, batch size (b) second. So x: (s,b,num_features), y,target_y: (s,b)
    # and style: Optional[(b,num_style_params)], style can be omitted or set to None, if it is not intended to be used.

    # For more references, see `priors/utils.py` for a pretty general implementation of a DataLoader
    # and `train.py` for the only call of it.



import random

import torch
from torch import nn


class ScaledDecoder(nn.Module):
    def __init__(self, ninp: int, nhid: int, nout: int) -> None:
        super().__init__()
        self.linear = nn.Linear(ninp, nhid)
        self.linear1 = nn.Linear(nhid, nout)
        self.linear2 = nn.Linear(nhid, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # return torch.cat([self.linear1(x), self.linear2(x)], -1)
        x = self.linear(x)
        x = nn.GELU()(x)
        temps = self.linear2(x).softmax(-1) @ torch.tensor(
            [1.0, 1.4, 1.7, 2.0, 5.0, 10.0, 20.0, 40.0, 80.0, 160.0], device=x.device
        )
        return self.linear1(x) / temps.unsqueeze(-1)


class FixedScaledDecoder(nn.Module):
    def __init__(self, ninp: int, nhid: int, nout: int) -> None:
        super().__init__()
        self.mapper = nn.Sequential(nn.Linear(ninp, nhid), nn.GELU(), nn.Linear(nhid, nout))
        self.T = nn.Parameter(torch.ones(10000) / 10000)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.mapper(x) / self.T.sum()



from __future__ import annotations

from collections.abc import Callable
import math
from typing import Any

import torch
import torch.nn as nn

from ifbo.utils import normalize_data


class StyleEncoder(nn.Module):
    def __init__(self, num_hyperparameters: int, em_size: int) -> None:
        super().__init__()
        self.em_size = em_size
        self.embedding = nn.Linear(num_hyperparameters, self.em_size)

    def forward(self, hyperparameters: dict[str, Any]) -> torch.Tensor:  # B x num_hps
        return self.embedding(hyperparameters)


class StyleEmbEncoder(nn.Module):
    def __init__(self, num_hyperparameters: int, em_size: int, num_embeddings: int = 100) -> None:
        super().__init__()
        assert num_hyperparameters == 1
        self.em_size = em_size
        self.embedding = nn.Embedding(num_embeddings, self.em_size)

    def forward(self, hyperparameters: torch.Tensor) -> torch.Tensor:  # B x num_hps
        return self.embedding(hyperparameters.squeeze(1))


class _PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.0) -> None:
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.d_model = d_model
        self.device_test_tensor = nn.Parameter(torch.tensor(1.0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # T x B x num_features
        assert self.d_model % x.shape[-1] * 2 == 0
        d_per_feature = self.d_model // x.shape[-1]
        pe = torch.zeros(*x.shape, d_per_feature, device=self.device_test_tensor.device)
        # position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        interval_size = 10
        div_term = (
            (1.0 / interval_size)
            * 2
            * math.pi
            * torch.exp(
                torch.arange(0, d_per_feature, 2, device=self.device_test_tensor.device).float()
                * math.log(math.sqrt(2))
            )
        )
        pe[..., 0::2] = torch.sin(x.unsqueeze(-1) * div_term)
        pe[..., 1::2] = torch.cos(x.unsqueeze(-1) * div_term)
        return self.dropout(pe).view(x.shape[0], x.shape[1], self.d_model)


class Positional(_PositionalEncoding):
    def __init__(self, num_features: int, emsize: int) -> None:
        super().__init__(d_model=emsize)
        self.num_features = num_features


class EmbeddingEncoder(nn.Module):
    def __init__(self, num_features: int, em_size: int, num_embs: int = 100) -> None:
        super().__init__()
        self.num_embs = num_embs
        self.embeddings = nn.Embedding(num_embs * num_features, em_size, max_norm=True)
        self.init_weights(0.1)
        self.min_max = (-2, +2)

    @property
    def width(self) -> float:
        return self.min_max[1] - self.min_max[0]

    def init_weights(self, initrange: float) -> None:
        self.embeddings.weight.data.uniform_(-initrange, initrange)

    def discretize(self, x: torch.Tensor) -> torch.Tensor:
        split_size = self.width / self.num_embs
        return (x - self.min_max[0] // split_size).int().clamp(0, self.num_embs - 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # T x B x num_features
        x_idxs = self.discretize(x)
        x_idxs += torch.arange(x.shape[-1], device=x.device).view(1, 1, -1) * self.num_embs
        return self.embeddings(x_idxs).mean(-2)


class Normalize(nn.Module):
    def __init__(self, mean: float, std: float) -> None:
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.mean) / self.std


class SqueezeBetween0and1(nn.Module):  # take care of test set here
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        width = x.max(0).values - x.min(0).values
        result = (x - x.min(0).values) / width
        result[(width == 0)[None].repeat(len(x), *[1] * (len(x.shape) - 1))] = 0.5
        return result


def get_normalized_uniform_encoder(
    encoder_creator: Callable[[int, int], nn.Module],
) -> Callable[[int, int], nn.Module]:
    """
    This can be used to wrap an encoder that is fed uniform samples in [0,1] and normalizes these to 0 mean and 1 std.
    For example, it can be used as `encoder_creator = get_normalized_uniform_encoder(encoders.Linear)`, now this can
    be initialized with `encoder_creator(feature_dim, in_dim)`.
    :param encoder:
    :return:
    """
    return lambda in_dim, out_dim: nn.Sequential(
        Normalize(0.5, math.sqrt(1 / 12)), encoder_creator(in_dim, out_dim)
    )


def get_normalized_encoder(
    encoder_creator: Callable[[int, int], nn.Module], data_std: float
) -> Callable[[int, int], nn.Module]:
    return lambda in_dim, out_dim: nn.Sequential(
        Normalize(0.0, data_std), encoder_creator(in_dim, out_dim)
    )


def get_log_dims(x: torch.Tensor, eps: float = 1e-10) -> torch.Tensor:
    logged_x = ((x + eps).log() - math.log(eps)) / (math.log(1.0 + eps) - math.log(eps))
    return logged_x


def add_log_neglog_dims(x: torch.Tensor, eps: float = 1e-10) -> torch.Tensor:
    logged_x = get_log_dims(x, eps) / 2.0
    neglogged_x = 1 - get_log_dims(1 - x, eps) / 2.0
    logged_x[x > 0.5] = neglogged_x[x > 0.5]
    return torch.stack([x, logged_x], -1).view(*x.shape[:-1], -1)


class AddLogNegLogDims(nn.Module):
    def __init__(self, eps: float = 1e-10) -> None:
        super().__init__()
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return add_log_neglog_dims(x, self.eps)


def get_logdim_encoder(
    encoder_creator: Callable[[int, int], nn.Module], eps: float = 1e-10
) -> Callable[[int, int], nn.Module]:
    return lambda in_dim, out_dim: nn.Sequential(
        AddLogNegLogDims(eps), encoder_creator(in_dim * 2, out_dim)
    )


class ZNormalize(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        std = x.std(-1, keepdim=True)
        std[std == 0.0] = 1.0
        return (x - x.mean(-1, keepdim=True)) / std


class ZNormalizePerDataset(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        std = x.std(0, keepdim=True)
        std[std == 0.0] = 1.0
        return (x - x.mean(0, keepdim=True)) / std


class AppendEmbeddingEncoder(nn.Module):
    def __init__(
        self, base_encoder: Callable[[torch.Tensor], torch.Tensor], num_features: int, emsize: int
    ) -> None:
        super().__init__()
        self.num_features = num_features
        self.base_encoder = base_encoder
        self.emb = nn.Parameter(torch.zeros(emsize))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if (x[-1] == 1.0).all():
            append_embedding = True
        else:
            assert (x[-1] == 0.0).all(), (
                "You need to specify as last position whether to append embedding. "
                "If you don't want this behavior, please use the wrapped encoder instead."
            )
            append_embedding = False
        x = x[:-1]
        encoded_x = self.base_encoder(x)
        if append_embedding:
            encoded_x = torch.cat(
                [encoded_x, self.emb[None, None, :].repeat(1, encoded_x.shape[1], 1)], 0
            )
        return encoded_x


def get_append_embedding_encoder(
    encoder_creator: Callable[[int, int], nn.Module],
) -> Callable[[int, int], nn.Module]:
    return lambda num_features, emsize: AppendEmbeddingEncoder(
        encoder_creator(num_features, emsize), num_features, emsize
    )


class VariableNumFeaturesEncoder(nn.Module):
    def __init__(
        self, base_encoder: Callable[[torch.Tensor], torch.Tensor], num_features: int
    ) -> None:
        super().__init__()
        self.base_encoder = base_encoder
        self.num_features = num_features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x * (self.num_features / x.shape[-1])
        x = torch.cat(
            (x, torch.zeros(*x.shape[:-1], self.num_features - x.shape[-1], device=x.device)), -1
        )
        return self.base_encoder(x)


def get_variable_num_features_encoder(
    encoder_creator: Callable[[int, int], nn.Module],
) -> Callable[[int, int], nn.Module]:
    return lambda num_features, emsize: VariableNumFeaturesEncoder(
        encoder_creator(num_features, emsize), num_features
    )


class NoMeanEncoder(nn.Module):
    """
    This can be useful for any prior that is translation invariant in x or y.
    A standard GP for example is translation invariant in x.
    That is, GP(x_test+const,x_train+const,y_train) = GP(x_test,x_train,y_train).
    """

    def __init__(self, base_encoder: Callable[[torch.Tensor], torch.Tensor]) -> None:
        super().__init__()
        self.base_encoder = base_encoder

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.base_encoder(x - x.mean(0, keepdim=True))


def get_no_mean_encoder(
    encoder_creator: Callable[[int, int], nn.Module],
) -> Callable[[int, int], nn.Module]:
    return lambda num_features, emsize: NoMeanEncoder(encoder_creator(num_features, emsize))


def MLP(num_features: int, emsize: int) -> nn.Sequential:
    return nn.Sequential(
        nn.Linear(num_features, emsize * 2),
        nn.ReLU(),
        nn.Linear(emsize * 2, emsize),
    )


class NanHandlingEncoder(nn.Module):
    def __init__(self, num_features: int, emsize: int, keep_nans: bool = True) -> None:
        super().__init__()
        self.num_features = 2 * num_features if keep_nans else num_features
        self.emsize = emsize
        self.keep_nans = keep_nans
        self.layer = nn.Linear(self.num_features, self.emsize)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.keep_nans:
            x = torch.cat(
                [
                    torch.nan_to_num(x, nan=0.0),
                    normalize_data(
                        torch.isnan(x) * -1
                        + torch.logical_and(torch.isinf(x), torch.sign(x) == 1) * 1
                        + torch.logical_and(torch.isinf(x), torch.sign(x) == -1) * 2
                    ),
                ],
                -1,
            )
        else:
            x = torch.nan_to_num(x, nan=0.0)
        return self.layer(x)


class Linear(nn.Linear):
    def __init__(self, num_features: int, emsize: int, replace_nan_by_zero: bool = False) -> None:
        super().__init__(num_features, emsize)
        self.num_features = num_features
        self.emsize = emsize
        self.replace_nan_by_zero = replace_nan_by_zero

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.replace_nan_by_zero:
            x = torch.nan_to_num(x, nan=0.0)
        return super().forward(x)

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        self.__dict__.setdefault("replace_nan_by_zero", True)


class Conv(nn.Module):
    def __init__(self, input_size: int, emsize: int) -> None:
        super().__init__()
        self.convs = torch.nn.ModuleList([nn.Conv2d(64 if i else 1, 64, 3) for i in range(5)])
        self.linear = nn.Linear(64, emsize)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        size = math.isqrt(x.shape[-1])
        assert size * size == x.shape[-1]
        x = x.reshape(*x.shape[:-1], 1, size, size)
        for conv in self.convs:
            if x.shape[-1] < 4:
                break
            x = conv(x)
            x.relu_()
        x = nn.AdaptiveAvgPool2d((1, 1))(x).squeeze(-1).squeeze(-1)
        return self.linear(x)


class CanEmb(nn.Embedding):
    def __init__(
        self, num_features: int, num_embeddings: int, embedding_dim: int, *args: Any, **kwargs: Any
    ) -> None:
        assert embedding_dim % num_features == 0
        embedding_dim = embedding_dim // num_features
        super().__init__(num_embeddings, embedding_dim, *args, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        lx = x.long()
        assert (lx == x).all(), "CanEmb only works with tensors of whole numbers"
        x = super().forward(lx)
        return x.view(*x.shape[:-2], -1)


def get_Canonical(num_classes: int) -> Callable[[int, int], nn.Module]:
    return lambda num_features, emsize: CanEmb(num_features, num_classes, emsize)


def get_Embedding(num_embs_per_feature: int = 100) -> Callable[[int, int], nn.Module]:
    return lambda num_features, emsize: EmbeddingEncoder(
        num_features, emsize, num_embs=num_embs_per_feature
    )



from __future__ import annotations

from functools import partial
from typing import Any

import torch
import torch.nn.functional as F
from torch.nn.modules.transformer import _get_activation_fn
from torch.nn.modules.transformer import Dropout
from torch.nn.modules.transformer import LayerNorm
from torch.nn.modules.transformer import Linear
from torch.nn.modules.transformer import Module
from torch.nn.modules.transformer import MultiheadAttention
from torch.nn.modules.transformer import Tensor
from torch.utils.checkpoint import checkpoint


class TransformerEncoderLayer(Module):
    r"""TransformerEncoderLayer is made up of self-attn and feedforward network.
    This standard encoder layer is based on the paper "Attention Is All You Need".
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of intermediate layer, relu or gelu (default=relu).
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False``.

    Examples::
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8)
        >>> src = torch.rand(10, 32, 512)
        >>> out = encoder_layer(src)

    Alternatively, when ``batch_first`` is ``True``:
        >>> encoder_layer = nn.TransformerEncoderLayer(d_model=512, nhead=8, batch_first=True)
        >>> src = torch.rand(32, 10, 512)
        >>> out = encoder_layer(src)
    """

    __constants__ = ["batch_first"]

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: str = "relu",
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        pre_norm: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        recompute_attn: bool = False,
        save_trainingset_representations: bool = False,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.self_attn = MultiheadAttention(
            d_model, nhead, dropout=dropout, batch_first=batch_first, **factory_kwargs
        )
        # Implementation of Feedforward model
        self.linear1 = Linear(d_model, dim_feedforward, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = Linear(dim_feedforward, d_model, **factory_kwargs)

        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps, **factory_kwargs)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)
        self.pre_norm = pre_norm
        self.recompute_attn = recompute_attn
        self.save_trainingset_representations = save_trainingset_representations
        self.saved_src_to_attend_to = None

        self.activation = _get_activation_fn(activation)

    def __setstate__(self, state: dict[str, Any]) -> None:
        if "activation" not in state:
            state["activation"] = F.relu
        super().__setstate__(state)
        self.__dict__.setdefault("save_trainingset_representations", False)

    def forward(
        self,
        src: Tensor,
        src_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
    ) -> Tensor:
        r"""Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        if self.save_trainingset_representations:
            assert (
                isinstance(src_mask, int) and not self.training
            ), "save_trainingset_representations is only supported in eval mode and requires src_mask to be an int"

        if self.pre_norm:
            src_ = self.norm1(src)
        else:
            src_ = src
        if isinstance(src_mask, tuple):
            # global attention setup
            assert not self.self_attn.batch_first
            assert src_key_padding_mask is None

            global_src_mask, trainset_src_mask, valset_src_mask = src_mask

            num_global_tokens = global_src_mask.shape[0]
            num_train_tokens = trainset_src_mask.shape[0]

            global_tokens_src = src_[:num_global_tokens]
            train_tokens_src = src_[num_global_tokens : num_global_tokens + num_train_tokens]
            global_and_train_tokens_src = src_[: num_global_tokens + num_train_tokens]
            eval_tokens_src = src_[num_global_tokens + num_train_tokens :]

            attn = partial(checkpoint, self.self_attn) if self.recompute_attn else self.self_attn

            global_tokens_src2 = attn(
                global_tokens_src,
                global_and_train_tokens_src,
                global_and_train_tokens_src,
                None,
                True,
                global_src_mask,
            )[0]
            train_tokens_src2 = attn(
                train_tokens_src,
                global_tokens_src,
                global_tokens_src,
                None,
                True,
                trainset_src_mask,
            )[0]
            eval_tokens_src2 = attn(eval_tokens_src, src_, src_, None, True, valset_src_mask)[0]

            src2 = torch.cat([global_tokens_src2, train_tokens_src2, eval_tokens_src2], dim=0)

        elif isinstance(src_mask, int):
            assert src_key_padding_mask is None
            single_eval_position = src_mask
            src_to_attend_to = src_[:single_eval_position]
            if self.save_trainingset_representations:
                if single_eval_position == src_.shape[0] or single_eval_position is None:
                    self.saved_src_to_attend_to = src_to_attend_to
                elif single_eval_position == 0:
                    if self.saved_src_to_attend_to is None:
                        raise ValueError(
                            "First save the trainingset representations by passing in a src_mask of None or the length of the src"
                        )
                    src_to_attend_to = self.saved_src_to_attend_to
                else:
                    raise ValueError(
                        "save_trainingset_representations only supports single_eval_position == 0 or single_eval_position == src.shape[0]"
                    )
            src_left = self.self_attn(
                src_[:single_eval_position],
                src_[:single_eval_position],
                src_[:single_eval_position],
            )[0]
            src_right = self.self_attn(
                src_[single_eval_position:], src_to_attend_to, src_to_attend_to
            )[0]
            src2 = torch.cat([src_left, src_right], dim=0)
        else:
            if self.recompute_attn:
                src2 = checkpoint(
                    self.self_attn, src_, src_, src_, src_key_padding_mask, True, src_mask
                )[0]
            else:
                src2 = self.self_attn(
                    src_, src_, src_, attn_mask=src_mask, key_padding_mask=src_key_padding_mask
                )[0]
        src = src + self.dropout1(src2)
        if not self.pre_norm:
            src = self.norm1(src)

        if self.pre_norm:
            src_ = self.norm2(src)
        else:
            src_ = src
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src_))))
        src = src + self.dropout2(src2)

        if not self.pre_norm:
            src = self.norm2(src)
        return src



from __future__ import annotations

from collections.abc import Callable
import math
from typing import Any

import torch
from torch import Tensor
import torch.nn as nn
from torch.nn import Module
from torch.nn import TransformerEncoder

from ifbo.layer import TransformerEncoderLayer
from ifbo.utils import bool_mask_to_att_mask
from ifbo.utils import SeqBN


class TransformerModel(nn.Module):
    def __init__(
        self,
        encoder: nn.Module,
        ninp: int,
        nhead: int,
        nhid: int,
        nlayers: int,
        dropout: float = 0.0,
        style_encoder: nn.Module | None = None,
        y_encoder: nn.Module | None = None,
        pos_encoder: nn.Module | None = None,
        decoder_dict: dict[str, tuple[nn.Module | None, int]] | None = None,
        input_normalization: bool = False,
        init_method: Callable | None = None,
        pre_norm: bool = False,
        activation: str = "gelu",
        recompute_attn: bool = False,
        num_global_att_tokens: int = 0,
        full_attention: bool = False,
        all_layers_same_init: bool = False,
        efficient_eval_masking: bool = True,
        decoder_once_dict: dict[str, tuple[nn.Module | None, int]] | None = None,
        return_all_outputs: bool = False,
        save_trainingset_representations: bool = False,
    ) -> None:
        super().__init__()
        self.model_type = "Transformer"

        def encoder_layer_creator() -> TransformerEncoderLayer:
            return TransformerEncoderLayer(
                ninp,
                nhead,
                nhid,
                dropout,
                activation=activation,
                pre_norm=pre_norm,
                recompute_attn=recompute_attn,
                save_trainingset_representations=save_trainingset_representations,
            )

        self.transformer_encoder = (
            TransformerEncoder(encoder_layer_creator(), nlayers)
            if all_layers_same_init
            else TransformerEncoderDiffInit(encoder_layer_creator, nlayers)  # type: ignore
        )
        self.ninp = ninp
        self.encoder = encoder
        self.y_encoder = y_encoder
        self.pos_encoder = pos_encoder
        self.return_all_outputs = return_all_outputs

        def make_decoder_dict(
            decoder_description_dict: dict[str, tuple[nn.Module | None, int]] | None,
        ) -> dict[str, nn.Module] | None:
            if decoder_description_dict is None or len(decoder_description_dict) == 0:
                return None
            initialized_decoder_dict = {}
            for decoder_key in decoder_description_dict:
                decoder_model, decoder_n_out = decoder_description_dict[decoder_key]
                if decoder_model is None:
                    initialized_decoder_dict[decoder_key] = nn.Sequential(
                        nn.Linear(ninp, nhid), nn.GELU(), nn.Linear(nhid, decoder_n_out)
                    )
                else:
                    initialized_decoder_dict[decoder_key] = decoder_model(
                        ninp, nhid, decoder_n_out
                    )
                print(
                    "Initialized decoder for",
                    decoder_key,
                    "with",
                    decoder_description_dict[decoder_key],
                    " and nout",
                    decoder_n_out,
                )
            return torch.nn.ModuleDict(initialized_decoder_dict)

        self.decoder_dict = make_decoder_dict(decoder_dict)
        self.decoder_dict_once = make_decoder_dict(decoder_once_dict)

        # N(0,1) is the initialization as the default of nn.Embedding
        self.decoder_dict_once_embeddings = (
            torch.nn.Parameter(torch.randn((len(self.decoder_dict_once), 1, ninp)))
            if self.decoder_dict_once is not None
            else None
        )
        # nn.Embedding(len(self.decoder_dict.keys()), nhid)
        self.input_ln = SeqBN(ninp) if input_normalization else None
        self.style_encoder = style_encoder
        self.init_method = init_method
        if num_global_att_tokens is not None:
            assert not full_attention
        self.global_att_embeddings = (
            nn.Embedding(num_global_att_tokens, ninp) if num_global_att_tokens else None
        )
        self.full_attention = full_attention
        self.efficient_eval_masking = efficient_eval_masking

        self.nhid = nhid

        self.init_weights()

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)
        self.__dict__.setdefault("efficient_eval_masking", False)
        if not hasattr(self, "decoder_dict_once"):
            self.__dict__.setdefault("decoder_dict_once", None)
        if hasattr(self, "decoder") and not hasattr(self, "decoder_dict"):
            self.add_module("decoder_dict", nn.ModuleDict({"standard": self.decoder}))
        self.__dict__.setdefault("return_all_outputs", False)

        def add_approximate_false(module: nn.Module) -> None:
            if isinstance(module, nn.GELU):
                module.__dict__.setdefault("approximate", "none")

        self.apply(add_approximate_false)

    @staticmethod
    def generate_square_subsequent_mask(sz: int) -> Tensor:
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_D_q_matrix(sz: int, query_size: int) -> Tensor:
        train_size = sz - query_size
        mask = torch.zeros(sz, sz) == 0
        mask[:, train_size:].zero_()
        mask |= torch.eye(sz) == 1
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_global_att_query_matrix(
        num_global_att_tokens: int, seq_len: int, num_query_tokens: int
    ) -> Tensor:
        train_size = seq_len + num_global_att_tokens - num_query_tokens
        sz = seq_len + num_global_att_tokens
        mask = torch.zeros(num_query_tokens, sz) == 0
        mask[:, train_size:].zero_()
        mask[:, train_size:] |= torch.eye(num_query_tokens) == 1
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_global_att_trainset_matrix(
        num_global_att_tokens: int, seq_len: int, num_query_tokens: int
    ) -> Tensor:
        trainset_size = seq_len - num_query_tokens
        mask = torch.zeros(trainset_size, num_global_att_tokens) == 0
        # mask[:,num_global_att_tokens:].zero_()
        # mask[:,num_global_att_tokens:] |= torch.eye(trainset_size) == 1
        return bool_mask_to_att_mask(mask)

    @staticmethod
    def generate_global_att_globaltokens_matrix(
        num_global_att_tokens: int, seq_len: int, num_query_tokens: int
    ) -> Tensor:
        mask = (
            torch.zeros(num_global_att_tokens, num_global_att_tokens + seq_len - num_query_tokens)
            == 0
        )
        return bool_mask_to_att_mask(mask)

    def init_weights(self) -> None:
        # initrange = 1.0
        # if isinstance(self.encoder,EmbeddingEncoder):
        #    self.encoder.weight.data.uniform_(-initrange, initrange)
        # self.decoder.bias.data.zero_()
        # self.decoder.weight.data.uniform_(-initrange, initrange)
        if self.init_method is not None:
            self.apply(self.init_method)
        for layer in self.transformer_encoder.layers:
            nn.init.zeros_(layer.linear2.weight)
            nn.init.zeros_(layer.linear2.bias)
            attns = (
                layer.self_attn
                if isinstance(layer.self_attn, nn.ModuleList)
                else [layer.self_attn]
            )
            for attn in attns:
                nn.init.zeros_(attn.out_proj.weight)
                nn.init.zeros_(attn.out_proj.bias)

    def forward(self, *args: Any, **kwargs: Any) -> Tensor:
        """
        This will perform a forward-pass (possibly recording gradients) of the model.
        We have multiple interfaces we support with this model:

        model(train_x, train_y, test_x, src_mask=None, style=None, only_return_standard_out=True)
        model((x,y), src_mask=None, single_eval_pos=None, only_return_standard_out=True)
        model((style,x,y), src_mask=None, single_eval_pos=None, only_return_standard_out=True)
        """
        if len(args) == 3:
            # case model(train_x, train_y, test_x, src_mask=None, style=None, only_return_standard_out=True)
            assert all(
                kwarg in {"src_mask", "style", "only_return_standard_out"}
                for kwarg in kwargs.keys()
            ), f"Unrecognized keyword argument in kwargs: {set(kwargs.keys()) - {'src_mask', 'style', 'only_return_standard_out'}}"
            x = args[0]
            if args[2] is not None:
                x = torch.cat((x, args[2]), dim=0)
            style = kwargs.pop("style", None)
            return self._forward((style, x, args[1]), single_eval_pos=len(args[0]), **kwargs)
        elif len(args) == 1 and isinstance(args, tuple):
            # case model((x,y), src_mask=None, single_eval_pos=None, only_return_standard_out=True)
            # case model((style,x,y), src_mask=None, single_eval_pos=None, only_return_standard_out=True)
            assert all(
                kwarg in {"src_mask", "single_eval_pos", "only_return_standard_out"}
                for kwarg in kwargs.keys()
            ), f"Unrecognized keyword argument in kwargs: {set(kwargs.keys()) - {'src_mask', 'single_eval_pos', 'only_return_standard_out'}}"
            return self._forward(*args, **kwargs)

    def _forward(
        self,
        src: tuple,
        src_mask: tuple | int | None = None,
        single_eval_pos: int | None = None,
        only_return_standard_out: bool = True,
    ) -> dict[str, Tensor] | tuple[dict[str, Tensor], dict[str, Tensor]]:
        assert isinstance(
            src, tuple
        ), "inputs (src) have to be given as (x,y) or (style,x,y) tuple"

        if len(src) == 2:  # (x,y) and no style
            src = (None,) + src

        style_src, x_src, y_src = src

        if single_eval_pos is None:
            single_eval_pos = x_src.shape[0]
        assert single_eval_pos is not None

        x_src = self.encoder(x_src)

        if self.decoder_dict_once is not None:
            x_src = torch.cat(
                [x_src, self.decoder_dict_once_embeddings.repeat(1, x_src.shape[1], 1)], dim=0
            )

        assert self.y_encoder is not None
        y_src = (
            self.y_encoder(y_src.unsqueeze(-1) if len(y_src.shape) < len(x_src.shape) else y_src)
            if y_src is not None
            else None
        )
        if self.style_encoder:
            assert style_src is not None, "style_src must be given if style_encoder is used"
            style_src = self.style_encoder(style_src).unsqueeze(0)
        else:
            style_src = torch.tensor([], device=x_src.device)
        global_src = (
            torch.tensor([], device=x_src.device)
            if self.global_att_embeddings is None
            else self.global_att_embeddings.weight.unsqueeze(1).repeat(1, x_src.shape[1], 1)
        )

        if src_mask is not None:
            assert self.global_att_embeddings is None or isinstance(src_mask, tuple)

        if src_mask is None:
            if self.global_att_embeddings is None:
                full_len = len(x_src) + len(style_src)
                if self.full_attention:
                    src_mask = bool_mask_to_att_mask(
                        torch.ones((full_len, full_len), dtype=torch.bool)
                    ).to(x_src.device)
                elif self.efficient_eval_masking:
                    src_mask = single_eval_pos + len(style_src)
                else:
                    src_mask = self.generate_D_q_matrix(full_len, len(x_src) - single_eval_pos).to(
                        x_src.device
                    )
            else:
                src_mask_args = (
                    self.global_att_embeddings.num_embeddings,
                    len(x_src) + len(style_src),
                    len(x_src) + len(style_src) - single_eval_pos,
                )
                src_mask = (
                    self.generate_global_att_globaltokens_matrix(*src_mask_args).to(x_src.device),
                    self.generate_global_att_trainset_matrix(*src_mask_args).to(x_src.device),
                    self.generate_global_att_query_matrix(*src_mask_args).to(x_src.device),
                )

        train_x = x_src[:single_eval_pos]
        if y_src is not None:
            train_x = train_x + y_src[:single_eval_pos]
        src = torch.cat([global_src, style_src, train_x, x_src[single_eval_pos:]], 0)

        if self.input_ln is not None:
            src = self.input_ln(src)

        if self.pos_encoder is not None:
            src = self.pos_encoder(src)

        output = self.transformer_encoder(src, src_mask)

        num_prefix_positions = len(style_src) + (
            self.global_att_embeddings.num_embeddings if self.global_att_embeddings else 0
        )
        if self.return_all_outputs:
            out_range_start = num_prefix_positions
        else:
            out_range_start = single_eval_pos + num_prefix_positions

        # In the line below, we use the indexing feature, that we have `x[i:None] == x[i:]`
        out_range_end = (
            -len(self.decoder_dict_once_embeddings) if self.decoder_dict_once is not None else None
        )

        # take care the output once are counted from the end
        output_once = (
            {k: v(output[-(i + 1)]) for i, (k, v) in enumerate(self.decoder_dict_once.items())}
            if self.decoder_dict_once is not None
            else {}
        )

        output = (
            {k: v(output[out_range_start:out_range_end]) for k, v in self.decoder_dict.items()}
            if self.decoder_dict is not None
            else {}
        )

        if only_return_standard_out:
            return output["standard"]

        if output_once:
            return output, output_once
        return output

    @torch.no_grad()
    def init_from_small_model(self, small_model: TransformerModel) -> None:
        assert (
            isinstance(self.decoder, nn.Linear)
            and isinstance(self.encoder, (nn.Linear, nn.Sequential))
            and isinstance(self.y_encoder, (nn.Linear, nn.Sequential))
        )

        def set_encoder_weights(
            my_encoder: nn.Linear | nn.Sequential,
            small_model_encoder: nn.Linear | nn.Sequential,
        ) -> None:
            my_encoder_linear, small_encoder_linear = (
                (my_encoder, small_model_encoder)
                if isinstance(my_encoder, nn.Linear)
                else (my_encoder[-1], small_model_encoder[-1])
            )
            small_in_dim = small_encoder_linear.out_features
            my_encoder_linear.weight.zero_()
            my_encoder_linear.bias.zero_()
            my_encoder_linear.weight[:small_in_dim] = small_encoder_linear.weight
            my_encoder_linear.bias[:small_in_dim] = small_encoder_linear.bias

        set_encoder_weights(self.encoder, small_model.encoder)
        set_encoder_weights(self.y_encoder, small_model.y_encoder)

        small_in_dim = small_model.decoder.in_features

        self.decoder.weight[:, :small_in_dim] = small_model.decoder.weight
        self.decoder.bias = small_model.decoder.bias

        for my_layer, small_layer in zip(
            self.transformer_encoder.layers, small_model.transformer_encoder.layers
        ):
            small_hid_dim = small_layer.linear1.out_features
            my_in_dim = my_layer.linear1.in_features

            # packed along q,k,v order in first dim
            my_in_proj_w = my_layer.self_attn.in_proj_weight
            small_in_proj_w = small_layer.self_attn.in_proj_weight

            my_in_proj_w.view(3, my_in_dim, my_in_dim)[:, :small_in_dim, :small_in_dim] = (
                small_in_proj_w.view(3, small_in_dim, small_in_dim)
            )
            my_layer.self_attn.in_proj_bias.view(3, my_in_dim)[:, :small_in_dim] = (
                small_layer.self_attn.in_proj_bias.view(3, small_in_dim)
            )

            my_layer.self_attn.out_proj.weight[:small_in_dim, :small_in_dim] = (
                small_layer.self_attn.out_proj.weight
            )
            my_layer.self_attn.out_proj.bias[:small_in_dim] = small_layer.self_attn.out_proj.bias

            my_layer.linear1.weight[:small_hid_dim, :small_in_dim] = small_layer.linear1.weight
            my_layer.linear1.bias[:small_hid_dim] = small_layer.linear1.bias

            my_layer.linear2.weight[:small_in_dim, :small_hid_dim] = small_layer.linear2.weight
            my_layer.linear2.bias[:small_in_dim] = small_layer.linear2.bias

            my_layer.norm1.weight[:small_in_dim] = (
                math.sqrt(small_in_dim / my_in_dim) * small_layer.norm1.weight
            )
            my_layer.norm2.weight[:small_in_dim] = (
                math.sqrt(small_in_dim / my_in_dim) * small_layer.norm2.weight
            )

            my_layer.norm1.bias[:small_in_dim] = small_layer.norm1.bias
            my_layer.norm2.bias[:small_in_dim] = small_layer.norm2.bias


class TransformerEncoderDiffInit(Module):
    r"""TransformerEncoder is a stack of N encoder layers

    Args:
        encoder_layer_creator: a function generating objects of TransformerEncoderLayer class without args (required).
        num_layers: the number of sub-encoder-layers in the encoder (required).
        norm: the layer normalization component (optional).
    """

    __constants__ = ["norm"]

    def __init__(
        self,
        encoder_layer_creator: Callable[[], nn.Module],
        num_layers: int,
        norm: nn.Module | None = None,
    ):
        super().__init__()
        self.layers = nn.ModuleList([encoder_layer_creator() for _ in range(num_layers)])
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        src: Tensor,
        mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
    ) -> Tensor:
        r"""Pass the input through the encoder layers in turn.

        Args:
            src: the sequence to the encoder (required).
            mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        output = src

        for mod in self.layers:
            output = mod(output, src_mask=mask, src_key_padding_mask=src_key_padding_mask)

        if self.norm is not None:
            output = self.norm(output)

        return output



from __future__ import annotations

from typing import Any

import torch

from ifbo.priors.prior import Batch
from ifbo.priors.utils import get_batch_to_dataloader
from ifbo.utils import default_device


def get_batch(
    batch_size: int,
    seq_len: int,
    num_features: int,
    hyperparameters: dict[str, Any],
    device: torch.device = default_device,
    batch_size_per_gp_sample: int | None = None,
    **kwargs: Any,
) -> Batch:
    batch_size_per_gp_sample = batch_size_per_gp_sample or (min(64, batch_size))
    num_models = batch_size // batch_size_per_gp_sample
    assert (
        num_models * batch_size_per_gp_sample == batch_size
    ), f"Batch size ({batch_size}) not divisible by batch_size_per_gp_sample ({batch_size_per_gp_sample})"

    args = {
        "device": device,
        "seq_len": seq_len,
        "num_features": num_features,
        "batch_size": batch_size_per_gp_sample,
    }

    prior_bag_priors_get_batch = hyperparameters["prior_bag_get_batch"]
    prior_bag_priors_p = [1.0] + [
        hyperparameters[f"prior_bag_exp_weights_{i}"]
        for i in range(1, len(prior_bag_priors_get_batch))
    ]

    weights = torch.tensor(prior_bag_priors_p, dtype=torch.float)  # create a tensor of weights
    batch_assignments = torch.multinomial(
        torch.softmax(weights, 0), num_models, replacement=True
    ).numpy()

    if "verbose" in hyperparameters and hyperparameters["verbose"]:
        print(
            "PRIOR_BAG:",
            weights,
            batch_assignments,
            num_models,
            batch_size_per_gp_sample,
            batch_size,
        )
        sample: list[Batch] = [
            prior_bag_priors_get_batch[int(prior_idx)](
                hyperparameters=hyperparameters, **args, **kwargs
            )
            for prior_idx in batch_assignments
        ]

    def merge(sample: list[Batch], k: str) -> Any:
        x = [getattr(x_, k) for x_ in sample]
        if torch.is_tensor(x[0]):
            return torch.cat(x, 1).detach()
        else:
            return [*x]

    merged_sample = {k: merge(sample, k) for k in sample[0].other_filled_attributes(set())}
    if hyperparameters.get("verbose"):
        print({k: v.shape for k, v in merged_sample.items()})

    return Batch(**merged_sample)


if __name__ == "__main__":
    DataLoader = get_batch_to_dataloader(get_batch)



from __future__ import annotations

import argparse
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Sequence
from dataclasses import dataclass
import datetime
import itertools
import math
import os
import random
from typing import Any

import numpy as np
import torch
from torch import nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR

from ifbo.bar_distribution import BarDistribution
from ifbo.priors.prior import Batch


@dataclass
class Curve:
    """
    A class to represent a performance curve.

    Attributes:
    -----------
    hyperparameters : torch.Tensor
        A tensor containing the hyperparameters. Should not have more than 10 dimensions and values should be in the range [0, 1].
    t : torch.Tensor
        A tensor containing the time steps. Values should be in the range [0, 1].
    y : torch.Tensor | None, optional
        A tensor containing the performance values (higher is better). Values should be in the range [0, 1]. Default is None.
    """

    hyperparameters: torch.Tensor
    t: torch.Tensor
    y: torch.Tensor | None = None


@dataclass(unsafe_hash=True)
class PredictionResult:
    """
    A dataclass for storing prediction results and computing various metrics.

    Attributes:
        logits (torch.Tensor): The logits output from the model.
        criterion (BarDistribution): The criterion used for computing various metrics.

    Methods:
        likelihood(y_test: torch.Tensor) -> torch.Tensor:
            Computes the negative log-likelihood of the test targets.

        ucb() -> torch.Tensor:
            Computes the upper confidence bound (UCB) of the logits.

        ei(y_best: torch.Tensor) -> torch.Tensor:
            Computes the expected improvement (EI) given the best observed value.

        pi(y_best: torch.Tensor) -> torch.Tensor:
            Computes the probability of improvement (PI) given the best observed value.

        quantile(q: float) -> torch.Tensor:
            Computes the quantile of the logits at the given quantile level.
    """

    logits: torch.Tensor
    criterion: BarDistribution

    def likelihood(self, y_test: torch.Tensor) -> torch.Tensor:
        """
        Computes the log-likelihood of the test targets.

        Args:
            y_test (torch.Tensor): The test targets.

        Returns:
            torch.Tensor: The log-likelihood.
        """
        return -self.criterion(self.logits, y_test).squeeze(1)

    def ucb(self) -> torch.Tensor:
        """
        Computes the upper confidence bound (UCB) of the logits.

        Returns:
            torch.Tensor: The upper confidence bound.
        """
        return self.criterion.ucb(self.logits, best_f=None).squeeze(1)

    def ei(self, y_best: torch.Tensor) -> torch.Tensor:
        """
        Computes the expected improvement (EI) given the best observed value.

        Args:
            y_best (torch.Tensor): The best observed value.

        Returns:
            torch.Tensor: The expected improvement.
        """
        return self.criterion.ei(self.logits, best_f=y_best).squeeze(1)

    def pi(self, y_best: torch.Tensor) -> torch.Tensor:
        """
        Computes the probability of improvement (PI) given the best observed value.

        Args:
            y_best (torch.Tensor): The best observed value.

        Returns:
            torch.Tensor: The probability of improvement.
        """
        return self.criterion.pi(self.logits, best_f=y_best).squeeze(1)

    def quantile(self, q: float) -> torch.Tensor:
        """
        Computes the quantile of the logits at the given quantile level.

        Args:
            q (float): The quantile level.

        Returns:
            torch.Tensor: The quantile at the given level.
        """
        return self.criterion.icdf(self.logits, q).squeeze(1)


# copied from huggingface
def get_cosine_schedule_with_warmup(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_training_steps: int,
    num_cycles: float = 0.5,
    last_epoch: int = -1,
) -> LambdaLR:
    """Create a schedule with a learning rate that decreases following the
    values of the cosine function between 0 and `pi * cycles` after a warmup
    period during which it increases linearly between 0 and 1.
    """

    def lr_lambda(current_step: int) -> float:
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(
            max(1, num_training_steps - num_warmup_steps)
        )
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return LambdaLR(optimizer, lr_lambda, last_epoch)


# copied from huggingface
def get_restarting_cosine_schedule_with_warmup(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_training_steps: int,
    steps_per_restart: int,
    num_cycles: float = 0.5,
    last_epoch: int = -1,
) -> LambdaLR:
    assert num_training_steps % steps_per_restart == 0

    def inner_lr_lambda(
        current_step: int, num_warmup_steps: int, num_training_steps: int
    ) -> float:
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(
            max(1, num_training_steps - num_warmup_steps)
        )
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    def lr_lambda(current_step: int) -> float:
        inner_step = current_step % steps_per_restart
        return inner_lr_lambda(
            inner_step,
            num_warmup_steps if current_step < steps_per_restart else 0,
            steps_per_restart,
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)


# copied from huggingface
def get_linear_schedule_with_warmup(
    optimizer: Optimizer,
    num_warmup_steps: int,
    num_training_steps: int,
    last_epoch: int = -1,
) -> LambdaLR:
    """
    Create a schedule with a learning rate that decreases linearly from the initial lr set in the optimizer to 0, after
    a warmup period during which it increases linearly from 0 to the initial lr set in the optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int) -> float:
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0,
            float(num_training_steps - current_step)
            / float(max(1, num_training_steps - num_warmup_steps)),
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def get_openai_lr(transformer_model: nn.Module) -> float:
    num_params = sum(p.numel() for p in transformer_model.parameters())
    return 0.003239 - 0.0001395 * math.log(num_params)


def get_weighted_single_eval_pos_sampler(
    max_len: int, min_len: int = 0, p: float = 1.0
) -> Callable[[], int]:
    """
    This gives a sampler that can be used for `single_eval_pos` which yields good performance for all positions p,
    where p <= `max_len`. At most `max_len` - 1 examples are shown to the Transformer.
    :return: Sampler that can be fed to `train()` as `single_eval_pos_gen`.
    """
    return lambda: random.choices(
        range(min_len, max_len),
        [1 / math.pow(((max_len - min_len) - i), p) for i in range(max_len - min_len)],
    )[0]


def get_uniform_single_eval_pos_sampler(max_len: int, min_len: int = 0) -> Callable[[], int]:
    """
    Just sample any evaluation position with the same weight
    :return: Sampler that can be fed to `train()` as `single_eval_pos_gen`.
    """
    return lambda: random.choices(range(min_len, max_len))[0]


class SeqBN(nn.Module):
    def __init__(self, d_model: int) -> None:
        super().__init__()
        self.bn = nn.BatchNorm1d(d_model)
        self.d_model = d_model

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert self.d_model == x.shape[-1]
        flat_x = x.view(-1, self.d_model)
        flat_x = self.bn(flat_x)
        return flat_x.view(*x.shape)


def set_locals_in_self(locals: dict[str, Any]) -> None:
    """
    Call this function like `set_locals_in_self(locals())` to set all local variables as object variables.
    Especially useful right at the beginning of `__init__`.
    :param locals: `locals()`
    """
    self = locals["self"]
    for var_name, val in locals.items():
        if var_name != "self":
            setattr(self, var_name, val)


default_device = "cuda:0" if torch.cuda.is_available() else "cpu:0"


# Copied from StackOverflow, but we do an eval on the values additionally
class StoredictKeyPair(argparse.Action):
    def __init__(
        self, option_strings: list[str], dest: str, nargs: int | None = None, **kwargs: Any
    ) -> None:
        self._nargs = nargs
        super(StoredictKeyPair, self).__init__(option_strings, dest, nargs=nargs, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        my_dict = {}
        if values is not None:
            for kv in values:
                k, v = kv.split("=")
                try:
                    my_dict[k] = eval(v)
                except NameError:
                    my_dict[k] = v
        setattr(namespace, self.dest, my_dict)


def get_nan_value(v: float, set_value_to_nan: float = 1.0) -> float:
    if random.random() < set_value_to_nan:
        return v
    else:
        return random.choice([-999, 0, 1, 999])


def to_ranking(data: torch.Tensor) -> torch.Tensor:
    x = data >= data.unsqueeze(-3)
    x = x.sum(0)
    return x


# TODO: Is there a better way to do this?
#   1. Cmparing to unique elements: When all values are different we still get quadratic blowup
#   2. Argsort(Argsort()) returns ranking, but with duplicate values there is an ordering which is problematic
#   3. Argsort(Argsort(Unique))->Scatter seems a bit complicated, doesn't have quadratic blowup, but how fast?
def to_ranking_low_mem(data: torch.Tensor) -> torch.Tensor:
    x = torch.zeros_like(data)
    for col in range(data.shape[-1]):
        x_ = data[:, :, col] >= data[:, :, col].unsqueeze(-2)
        x_ = x_.sum(0)
        x[:, :, col] = x_
    return x


def nan_handling_missing_for_unknown_reason_value(nan_prob: float = 1.0) -> float:
    return get_nan_value(float("nan"), nan_prob)


def nan_handling_missing_for_no_reason_value(nan_prob: float = 1.0) -> float:
    return get_nan_value(float("-inf"), nan_prob)


def nan_handling_missing_for_a_reason_value(nan_prob: float = 1.0) -> float:
    return get_nan_value(float("inf"), nan_prob)


def torch_nanmean(x: torch.Tensor, axis: int = 0, return_nanshare: bool = False) -> torch.Tensor:
    num = torch.where(torch.isnan(x), torch.full_like(x, 0), torch.full_like(x, 1)).sum(axis=axis)
    value = torch.where(torch.isnan(x), torch.full_like(x, 0), x).sum(axis=axis)
    if return_nanshare:
        return value / num, 1.0 - num / x.shape[axis]
    return value / num


def torch_nanstd(x: torch.Tensor, axis: int = 0) -> torch.Tensor:
    num = torch.where(torch.isnan(x), torch.full_like(x, 0), torch.full_like(x, 1)).sum(axis=axis)
    value = torch.where(torch.isnan(x), torch.full_like(x, 0), x).sum(axis=axis)
    mean = value / num
    mean_broadcast = torch.repeat_interleave(mean.unsqueeze(axis), x.shape[axis], dim=axis)
    return torch.sqrt(torch.nansum(torch.square(mean_broadcast - x), axis=axis) / (num - 1))


def normalize_data(
    data: torch.Tensor, normalize_positions: int = -1, return_scaling: bool = False
) -> torch.Tensor | tuple[torch.Tensor, tuple[float, float]]:
    if normalize_positions > 0:
        mean = torch_nanmean(data[:normalize_positions], axis=0)
        std = torch_nanstd(data[:normalize_positions], axis=0) + 0.000001
    else:
        mean = torch_nanmean(data, axis=0)
        std = torch_nanstd(data, axis=0) + 0.000001
    data = (data - mean) / std
    data = torch.clip(data, min=-100, max=100)

    if return_scaling:
        return data, (mean, std)
    return data


def remove_outliers(
    X: torch.Tensor, n_sigma: int = 4, normalize_positions: int = -1
) -> torch.Tensor:
    # Expects T, B, H
    assert len(X.shape) == 3, "X must be T,B,H"
    # for b in range(X.shape[1]):
    # for col in range(X.shape[2]):
    data = X if normalize_positions == -1 else X[:normalize_positions]
    data_clean = data[:].clone()
    data_mean, data_std = torch_nanmean(data, axis=0), torch_nanstd(data, axis=0)
    cut_off = data_std * n_sigma
    lower, upper = data_mean - cut_off, data_mean + cut_off

    data_clean[torch.logical_or(data_clean > upper, data_clean < lower)] = np.nan
    data_mean, data_std = (
        torch_nanmean(data_clean, axis=0),
        torch_nanstd(data_clean, axis=0),
    )
    cut_off = data_std * n_sigma
    lower, upper = data_mean - cut_off, data_mean + cut_off

    X = torch.maximum(-torch.log(1 + torch.abs(X)) + lower, X)
    X = torch.minimum(torch.log(1 + torch.abs(X)) + upper, X)
    return X


def bool_mask_to_att_mask(mask: torch.Tensor) -> torch.Tensor:
    return mask.float().masked_fill(mask == 0, float("-inf")).masked_fill(mask == 1, float(0.0))


def print_on_master_only(is_master: bool) -> None:
    import builtins as __builtin__

    builtin_print = __builtin__.print

    def print(*args: Any, **kwargs: Any) -> None:
        force = kwargs.pop("force", False)
        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def init_dist(device: torch.device) -> tuple[bool, int, torch.device]:
    print("init dist")
    if "LOCAL_RANK" in os.environ:
        # launched with torch.distributed.launch
        rank = int(os.environ["LOCAL_RANK"])
        print("torch.distributed.launch and my rank is", rank)
        torch.cuda.set_device(rank)
        os.environ["CUDA_VISIBLE_DEVICES"] = str(rank)
        torch.distributed.init_process_group(
            backend="nccl",
            init_method="env://",
            timeout=datetime.timedelta(seconds=20),
            world_size=torch.cuda.device_count(),
            rank=rank,
        )
        torch.distributed.barrier()
        print_on_master_only(rank == 0)
        print(
            f"Distributed training on {torch.cuda.device_count()} GPUs, this is rank {rank}, "
            "only I can print, but when using print(..., force=True) it will print on all ranks."
        )
        return True, rank, f"cuda:{rank}"
    elif "SLURM_PROCID" in os.environ and torch.cuda.device_count() > 1:
        # this is for multi gpu when starting with submitit
        assert device != "cpu:0"
        rank = int(os.environ["SLURM_PROCID"])
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "12355"
        torch.cuda.set_device(rank)
        os.environ["CUDA_VISIBLE_DEVICES"] = str(rank)
        print("distributed submitit launch and my rank is", rank)
        torch.distributed.init_process_group(
            backend="nccl",
            init_method="env://",
            timeout=datetime.timedelta(seconds=20),
            world_size=torch.cuda.device_count(),
            rank=rank,
        )
        torch.distributed.barrier()
        print_on_master_only(rank == 0)
        print(
            f"Distributed training on {torch.cuda.device_count()} GPUs, this is rank {rank}, "
            "only I can print, but when using print(..., force=True) it will print on all ranks."
        )

        return True, rank, f"cuda:{rank}"
    else:
        print("Not using distributed")
        # will not change any of the behavior of print, but allows putting the force=True in the print calls
        print_on_master_only(True)
        return False, 0, device


# NOP decorator for python with statements (x = NOP(); with x:)
class NOP:
    def __enter__(self) -> None:
        pass

    def __exit__(self, type: Any, value: Any, traceback: Any) -> None:
        pass


def check_compatibility(dl: torch.utils.data.DataLoader) -> None:
    if hasattr(dl, "num_outputs"):
        print("`num_outputs` for the DataLoader is deprecated. It is assumed to be 1 from now on.")
        assert dl.num_outputs != 1, (
            "We assume num_outputs to be 1. Instead of the num_ouputs change your loss."
            "We specify the number of classes in the CE loss."
        )


def product_dict(dic: dict[str, Any]) -> Generator[dict[str, Any], None, None]:
    keys = dic.keys()
    vals = dic.values()
    for instance in itertools.product(*vals):
        yield dict(zip(keys, instance))


def to_tensor(x: torch.Tensor, device: torch.device | None = None) -> torch.Tensor:
    if isinstance(x, torch.Tensor):
        return x.to(device)
    else:
        return torch.tensor(x, device=device)


printed_already = set()


def print_once(*msgs: str) -> None:
    msg = " ".join([repr(m) for m in msgs])
    if msg not in printed_already:
        print(msg)
        printed_already.add(msg)


def tokenize(
    context: list[Curve], query: list[Curve], device: torch.device | None = None
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    # takes as input a list of curves and query points (does not have y values)
    # returns the tokenized representation of
    #   - context curves: ([id curve, t value, hyperparameters]) and the corresponding y values.
    #   - query points: ([id curve, t value, hyperparameters])
    # The id curve is a unique identifier for each curve in the context.

    config_to_id: dict[torch.Tensor, int] = {}
    context_tokens = []
    context_y_values = []
    query_tokens = []
    current_id = 1

    def get_curve_id(hyperparameters: torch.Tensor) -> int:
        nonlocal current_id
        for config, cid in config_to_id.items():
            if torch.equal(config, hyperparameters):
                return cid
        config_to_id[hyperparameters] = current_id
        current_id += 1
        return config_to_id[hyperparameters]

    for curve in context:
        curve_id = get_curve_id(curve.hyperparameters)
        num_points = curve.t.size(0)
        for i in range(num_points):
            context_tokens.append(
                torch.cat(
                    (torch.tensor([curve_id, curve.t[i].item()]), curve.hyperparameters.cpu())
                )
            )
            assert curve.y is not None
            context_y_values.append(curve.y[i])

    for curve in query:
        curve_id = get_curve_id(curve.hyperparameters)
        num_points = curve.t.size(0)
        for i in range(num_points):
            query_tokens.append(
                torch.cat(
                    (torch.tensor([curve_id, curve.t[i].item()]), curve.hyperparameters.cpu())
                )
            )

    # Convert lists to tensors
    context_tokens_tensor = torch.stack(context_tokens, dim=0).to(device)
    context_y_values_tensor = torch.stack(context_y_values, dim=0).to(device)
    query_tokens_tensor = torch.stack(query_tokens, dim=0).to(device)

    return context_tokens_tensor, context_y_values_tensor, query_tokens_tensor


def detokenize(
    batch: Batch, context_size: int, device: torch.device | None = None
) -> tuple[list[Curve], list[Curve]]:
    (
        context_tokens_tensor,
        context_y_values_tensor,
        query_tokens_tensor,
        query_y_values_tensor,
    ) = (
        batch.x.squeeze(1)[:context_size, ...],
        batch.y.squeeze(1)[:context_size, ...],
        batch.x.squeeze(1)[context_size:, ...],
        batch.y.squeeze(1)[context_size:, ...],
    )
    id_to_config: dict[int, torch.Tensor] = {}
    context_curves: dict[int, list[tuple[float, float]]] = {}
    query_curves: dict[int, list[tuple[float, float]]] = {}
    used_ids: set[int] = set()
    all_possible_ids: set[int] = set(range(1, 1001))

    def get_curve_config(curve_id: int) -> torch.Tensor:
        if curve_id in id_to_config:
            return id_to_config[curve_id]
        else:
            raise KeyError(f"Curve ID {curve_id} not found in id_to_config")

    # Process context tokens and y values
    for i in range(context_tokens_tensor.size(0)):
        token = context_tokens_tensor[i]
        y_value = context_y_values_tensor[i]

        curve_id = int(token[0].item())
        x_value = token[1].item()
        configuration = token[2:]

        if curve_id not in id_to_config:
            id_to_config[curve_id] = configuration
            used_ids.add(curve_id)

        if curve_id not in context_curves:
            context_curves[curve_id] = []

        context_curves[curve_id].append((x_value, y_value.item()))

    unused_ids = all_possible_ids - used_ids

    # Process query tokens
    for i in range(query_tokens_tensor.size(0)):
        token = query_tokens_tensor[i]
        y_value = (
            query_y_values_tensor[i]
            if query_y_values_tensor is not None
            else torch.tensor([0.0]) * float("nan")
        )

        curve_id = int(token[0].item())
        x_value = token[1].item()
        configuration = token[2:]

        # Assign a new unique ID for configurations with curve_id 0 not in context tokens
        if curve_id not in used_ids:
            found = False
            for existing_id, config in id_to_config.items():
                if torch.equal(config, configuration):
                    curve_id = existing_id
                    found = True
                    break
            if not found:
                if not unused_ids:
                    raise ValueError("No unused IDs available")
                curve_id = unused_ids.pop()
                id_to_config[curve_id] = configuration

        if curve_id not in query_curves:
            query_curves[curve_id] = []

        query_curves[curve_id].append((x_value, y_value.item()))

    # Convert the context curves dictionary to list of Curve objects
    context_list = []
    for curve_id, points in context_curves.items():
        x_values = torch.tensor([p[0] for p in points]).to(device)
        y_values = torch.tensor([p[1] for p in points]).to(device)
        configuration = get_curve_config(curve_id)
        context_list.append(Curve(t=x_values, y=y_values, hyperparameters=configuration))

    # Convert the query curves dictionary to list of Curve objects
    query_list = []
    for curve_id, points in query_curves.items():
        x_values = torch.tensor([p[0] for p in points]).to(device)
        if query_y_values_tensor is not None:
            y_values = torch.tensor([p[1] for p in points]).to(device)
        configuration = get_curve_config(curve_id)
        query_list.append(
            Curve(
                t=x_values,
                hyperparameters=configuration,
                y=y_values if query_y_values_tensor is not None else None,
            )
        )

    return context_list, query_list



import ifbo
import unittest


class TestSurrogateModel(unittest.TestCase):

    def setUp(self):
        # Initialize the surrogate model and any required data
        self.model = ifbo.surrogate.FTPFN(version="0.0.1")
        single_eval_pos = 700 
        batch = ifbo.priors.ftpfn_prior.get_batch(
            batch_size=1,
            seq_len=1000,
            num_features=12,
            single_eval_pos=single_eval_pos 
        )
        self.context, self.query = ifbo.utils.detokenize(batch, context_size=single_eval_pos, device="cpu")

    def test_prediction_shape(self):
        # Test if the prediction output has the correct shape
        predictions = self.model.predict(self.context, self.query)
        self.assertEqual(len(predictions), len(self.query))
    
    def assertBetween(self, value, min, max):
        """Fail if value is not between min and max (inclusive)."""
        self.assertTrue((value >= min).all())
        self.assertTrue((value <= max).all())

    def test_prediction_values(self):
        predictions = self.model.predict(self.context, self.query)
        for prediction in predictions:
            for q in [0.01, 0.5, 0.99]:
                self.assertBetween(prediction.quantile(q), 0, 1)
            self.assertBetween(prediction.ucb(), 0, 1)
            self.assertBetween(prediction.ei(0.5), 0, 1)
            self.assertBetween(prediction.pi(0.5), 0, 1)
    
    def test_exception_hyperparameters(self):
        """Test if the model raises an exception for invalid input."""
        invalid_context = self.context.copy()
        invalid_context[0].hyperparameters[-1] = -1
        self.assertRaises(Exception, self.model.predict, invalid_context, self.query)

        invalid_context = self.context.copy()
        invalid_context[0].hyperparameters[-1] = 2
        self.assertRaises(Exception, self.model.predict, invalid_context, self.query)

        invalid_query = self.query.copy()
        invalid_query[0].hyperparameters[-1] = -1
        self.assertRaises(Exception, self.model.predict, self.context, invalid_query)

        invalid_query = self.query.copy()
        invalid_query[0].hyperparameters[-1] = 2
        self.assertRaises(Exception, self.model.predict, self.context, invalid_query)
    
    def test_exception_step(self):
        """Test if the model raises an exception for invalid input."""
        invalid_context = self.context.copy()
        invalid_context[0].t = -1
        self.assertRaises(Exception, self.model.predict, invalid_context, self.query)

        invalid_context = self.context.copy()
        invalid_context[0].t = 2
        self.assertRaises(Exception, self.model.predict, invalid_context, self.query)

        invalid_query = self.query.copy()
        invalid_query[0].t = -1
        self.assertRaises(Exception, self.model.predict, self.context, invalid_query)

        invalid_query = self.query.copy()
        invalid_query[0].t = 2
        self.assertRaises(Exception, self.model.predict, self.context, invalid_query)
    
    def test_exception_performance(self):
        """Test if the model raises an exception for invalid input."""
        invalid_context = self.context.copy()
        invalid_context[0].y = -1
        self.assertRaises(Exception, self.model.predict, invalid_context, self.query)

        invalid_context = self.context.copy()
        invalid_context[0].y = 2
        self.assertRaises(Exception, self.model.predict, invalid_context, self.query)

        invalid_query = self.query.copy()
        invalid_query[0].y = -1
        self.assertRaises(Exception, self.model.predict, self.context, invalid_query)

        invalid_query = self.query.copy()
        invalid_query[0].y = 2
        self.assertRaises(Exception, self.model.predict, self.context, invalid_query)

        

if __name__ == '__main__':
    unittest.main()


import argparse
import os
from pathlib import Path
import tarfile

import requests


VERSION_MAP = {
    "0.0.1": dict(
        url="https://ml.informatik.uni-freiburg.de/research-artifacts/ifbo/ftpfnv0.0.1.tar.gz",
        name="ftpfnv0.0.1",
        final_name="bopfn_broken_unisep_1000curves_10params_2M",
        extension="pt",
    )
}


# Helper functions to generate the file names
def FILENAME(version: str) -> str:
    return f"{VERSION_MAP[version].get('name')}.tar.gz"


def FILE_URL(version: str) -> str:
    return f"{VERSION_MAP[version].get('url')}"


def WEIGHTS_FILE_NAME(version: str) -> str:
    return f"{VERSION_MAP[version].get('name')}.{VERSION_MAP[version].get('extension')}"


def WEIGHTS_FINAL_NAME(version: str) -> str:
    return f"{VERSION_MAP[version].get('final_name')}.{VERSION_MAP[version].get('extension')}"


def download_and_decompress(url: str, path: Path) -> None:
    """Helper function to download a file from a URL and decompress it and store by given name.

    Args:
        url (str): URL of the file to download
        path (Path): Path along with filename to save the downloaded file

    Returns:
        bool: Flag to indicate if the download and decompression was successful
    """
    # Check if the file already exists
    if path.exists():
        return

    # Send a HTTP request to the URL of the file
    response = requests.get(url, allow_redirects=True)

    # Check if the request is successful
    if response.status_code != 200:
        raise ValueError(
            f"Failed to download the surrogate from {url}."
            f" Recieved HTTP status code: {response.status_code}."
            "Please either try again later, use an alternative link or contact the authors through github."
        )

    # Save the .tar.gz file
    with open(path, "wb") as f:
        f.write(response.content)

    # Decompress the .tar.gz file
    with tarfile.open(path, "r:gz") as tar:
        tar.extractall(path.parent.absolute())


def parse_args() -> argparse.Namespace:
    """Helper function to parse the command line arguments."""
    args = argparse.ArgumentParser()

    args.add_argument(
        "--version", type=str, default="0.0.1", help="The version of the PFN model to download"
    )
    args.add_argument(
        "--path", type=str, default=None, help="The path to save the downloaded file"
    )

    parser = args.parse_args()
    return parser


if __name__ == "__main__":
    args = parse_args()

    assert args.version in VERSION_MAP, "The version provided is not available"

    if args.path is None:
        args.path = Path(__file__).parent.absolute() / ".." / ".." / "PFNS4HPO" / "final_models"
    else:
        args.path = Path(args.path)

    if not args.path.exists():
        os.makedirs(args.path)

    # Use the function
    download_and_decompress(
        url=VERSION_MAP[args.version]["url"], path=args.path / FILENAME(args.version)
    )
    print(f"Successfully downloaded FT-PFN v{args.version} in to {args.path}!")
